/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton@ranton.org

For custom software development and consulting please visit http://www.madlogic.com .

Copyright (c) 2003, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <windows.h>
#include <shlobj.h>
#include <intshcut.h>
#include <tchar.h>
#include <cstddef>
#include <crtdbg.h> // atlconv.h is stupid and uses this without including it.
#include <atlconv.h> // string conversion stuff provided by ATL
#include <fstream>

#include "resource.h"

using namespace std;


const char *kAppName = "Fav";


// function prototypes.
long FAR PASCAL MainWindowProc( HWND hwnd, UINT msg, UINT wParam, LONG lParam );
BOOL CALLBACK MainDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam);
HRESULT GetURL(const TCHAR *filePathStr, TCHAR **outURLStr);
BOOL UpdateDialog( HWND hwndDlg, LPCTSTR nameStr, LPCTSTR urlStr);
UINT APIENTRY OFNHookProc( HWND hwnd,UINT msg, WPARAM wParam, LPARAM lParam );
BOOL GetHTMLSaveFileName(HINSTANCE inInstance, char *filename, unsigned buffsize);
char *GetFileExtension(char *inFileName);
void ProcessFavorites(HWND hwndDlg, const char *outputPath);
HRESULT ProcessFavFolder(HWND hwndDlg, IMalloc *pMalloc, IShellFolder* psfFolder, ostream& file);


// Globals
BOOL gScanning = FALSE, gCancelled = FALSE;
HINSTANCE gInstance = NULL;
HWND gDialog = NULL;

// Main
int APIENTRY WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPTSTR lpCmdLine,
	int nCmdShow)
{
	HWND mainWindow = NULL;
	HWND mainDialog = NULL;
	WNDCLASS wc;
	
	gInstance = hInstance;

	// init COM
	CoInitialize(NULL);

	// register our window class
	wc.style = CS_HREDRAW | CS_VREDRAW; // what are these flags?
	wc.lpfnWndProc = MainWindowProc;

	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;

	wc.hIcon = LoadIcon(NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor(NULL, IDC_ARROW );
	wc.hbrBackground = (HBRUSH)GetStockObject(  WHITE_BRUSH );

	// no menu for this one.
	wc.lpszMenuName = NULL; // MAKEINTRESOURCE( IDR_MENU1 );
	wc.lpszClassName = kAppName;

	if( !RegisterClass( &wc ) )
	{
		MessageBox(NULL,"Unable to register window class","Error",MB_OK | MB_ICONERROR );
		goto cleanup;
	}

	// create window 
	mainWindow = CreateWindow( 
		kAppName, // window class name
		kAppName, // caption 
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN ,      // window style            
		CW_USEDEFAULT,            // initial X position      
		CW_USEDEFAULT,            // initial Y position      
		CW_USEDEFAULT,            // initial X size          
		CW_USEDEFAULT,            // initial Y size    
		NULL,                     // parent window handle    
		NULL,                     // set the window menu handle  		
		hInstance,                // program instance handle
		NULL  );                  // creation parameters   

	if( !mainWindow )
	{
		MessageBox(NULL,"Error creating window","Error",MB_OK | MB_ICONERROR );
		goto cleanup;
	}

	// Load dialog into window
	mainDialog = CreateDialog(hInstance,MAKEINTRESOURCE(IDD_FAV),mainWindow,MainDialogProc);
	if( !mainDialog )
	{
		MessageBox(mainWindow,"Error creating dialog","Error",MB_OK | MB_ICONERROR );
		goto cleanup;
	}
	gDialog = mainDialog;

	// set to appropriate sizes.
	RECT r;
	GetClientRect(mainDialog,&r);
	// shouldn't really have this hardcoded.
	SetWindowPos(mainWindow,NULL,0,0,r.right-r.left,r.bottom-r.top+20,SWP_NOZORDER | SWP_NOMOVE);

	GetClientRect(mainWindow,&r);
	SetWindowPos(mainDialog,NULL,0,0,r.right-r.left,r.bottom-r.top,SWP_NOZORDER | SWP_NOMOVE);


	// show the windows.
	ShowWindow( mainDialog, true );
	SetFocus( mainDialog );
	ShowWindow( mainWindow, true );
	

	// Message loop 
	BOOL bRet;
	MSG msg;

	while ( (bRet = GetMessage(&msg, NULL, 0, 0)) != 0 ) 
	{ 
		if (bRet == -1 )
		{
			// handle the error and possibly exit
		}
		else 
		{
			if (!IsWindow(mainDialog) || !IsDialogMessage(mainDialog, &msg)) 
			{ 
				TranslateMessage(&msg); 
				DispatchMessage(&msg); 
			}
		}
	}

cleanup:
	CoUninitialize();

	return 0;
}






// this is the Window Proc for our main window
long FAR PASCAL MainWindowProc( HWND hwnd, UINT msg, UINT wParam, LONG lParam )
{
	HDC hdc;
	PAINTSTRUCT ps;
	RECT r;
			
	switch( msg )
	{
		case WM_CREATE:
			break;

		case WM_ACTIVATE:
			break;
		
		case WM_PAINT:
			// we don't have anything we have to paint by hand from this window proc.
			hdc = BeginPaint(hwnd,&ps);

			// nothing to draw by hand.

			EndPaint(hwnd,&ps);

			return 0;
		case WM_SIZE:
			GetClientRect(hwnd,&r);
			SetWindowPos(gDialog,NULL,0,0,r.right-r.left,r.bottom-r.top,SWP_NOZORDER | SWP_NOMOVE);
	
			break;
		case WM_CLOSE:
			PostQuitMessage(0);
			break;
		case WM_DESTROY:
			PostQuitMessage(0);
			break;
		default:
			return DefWindowProc(hwnd,msg,wParam,lParam) ;
	}

	return 0;
}

// dialog hook for the open dialog
// using this hook to make sure the save dialog is centered correctly
UINT APIENTRY OFNHookProc( HWND hwnd,UINT msg, WPARAM wParam, LPARAM lParam )
{
	if(msg == WM_INITDIALOG)
	{
		// to get this thing centered in the screen, get the parent of the passed window
		hwnd = GetParent(hwnd);

		RECT parentRect;
		RECT windowRect;


		HWND parent = GetParent(hwnd);

		// get the dimensions of this dialog
		GetWindowRect(hwnd, &windowRect);

		// get the dimensions of the parent
		if(!parent)
		{
			parentRect.top = 0;
			parentRect.left = 0;
			parentRect.bottom = GetSystemMetrics(SM_CYFULLSCREEN);
			parentRect.right = GetSystemMetrics(SM_CXFULLSCREEN);
		}
		else
		{
			GetWindowRect(GetParent(hwnd), &parentRect);
		}

		int x = 0, y = 0;

		if(parentRect.right - parentRect.left > windowRect.right - windowRect.left)
			x = parentRect.right - parentRect.left - (windowRect.right - windowRect.left);

		if(parentRect.bottom - parentRect.top > windowRect.bottom - windowRect.top)
			y = parentRect.bottom - parentRect.top - (windowRect.bottom - windowRect.top);


		// center horizontally
		x = x / 2;
		// want the dialog more skewed toward the top
		y = y / 3;

		// set the new position to center the window
		SetWindowPos(hwnd, NULL, x,y,0,0, SWP_NOZORDER | SWP_NOSIZE);
	}
	return false; // false indicated continue processing to the default handler
}

// returns pointer to null terminator of string if not found
// if found, returns pointer to '.'
char *GetFileExtension(char *inFileName)
{
	char *dot = NULL;


	while(*inFileName)
	{
		switch(*inFileName)
		{
		case ':':
		case '\\':
		case '//':
			// if there was a dot already, it had nothing to do with the
			// filename, and something to do with a folder name or something
			dot = NULL; 
			break;
		case '.':
			dot = inFileName;
		default:
			break;
		}
		inFileName++;
	}

	if(!dot)
		dot = inFileName + strlen(inFileName);

	return dot;

}

// get the file name to save it to
// false if user hit cancel or dialog failed
BOOL GetHTMLSaveFileName(HINSTANCE inInstance, char *filename, unsigned buffsize)
{
	OPENFILENAME ofn;

	char myfilter[] = "HTML Files\0*.HTML;*.HTM\0";

	filename[0] = 0;

	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = GetDesktopWindow();
	ofn.hInstance = inInstance;
	ofn.lpstrFilter = myfilter;

	ofn.lpstrCustomFilter = NULL;  // for custom filters
	ofn.nMaxCustFilter = 0;        // ignored

	// set index to the last used file type
	ofn.nFilterIndex = 1;

	ofn.lpstrFile = filename;
	ofn.nMaxFile = buffsize;

    ofn.lpstrFileTitle = NULL; // if we want the filename without full path
    ofn.nMaxFileTitle = 0; // ignored

    ofn.lpstrInitialDir = NULL; // tells system to use current directory
    ofn.lpstrTitle = NULL; // title of the dialog box, null for default

    ofn.Flags = OFN_HIDEREADONLY | OFN_ENABLEHOOK | OFN_EXPLORER | OFN_OVERWRITEPROMPT;
    ofn.nFileOffset = 0; //
    ofn.nFileExtension = 0; //
    ofn.lpstrDefExt = NULL;// i can put a default extension here, or I can do it after the fact or in a hook
    ofn.lCustData = NULL;
    ofn.lpfnHook = OFNHookProc; // put in our hook
    ofn.lpTemplateName = NULL;

	if(GetSaveFileName( &ofn))
	{
		char *ext = GetFileExtension(filename);
		if(!(*ext) &&  // look to see if there is no extension
			(strlen(filename) + 5 < buffsize)  // make sure there is room to attach an extension
			)
		{
			ext = ".HTM";
			strcat(filename, ext);
		}
		return true;
	}
	else
	{
		return false;
	}
}


BOOL CALLBACK MainDialogProc(HWND hwndDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	char gOutputPath[MAX_PATH];

    switch (message) 
    { 
        case WM_INITDIALOG: 
            return TRUE; 
 
        case WM_COMMAND: 
			// handle button presses.
            switch (LOWORD(wParam)) 
            { 	
			case IDC_START:
				// Start scanning favorites
				if( !gScanning ) 
				{
					// select an output location
					if( GetHTMLSaveFileName(gInstance,gOutputPath,MAX_PATH) )
					{
						// and process the favorites.
						ProcessFavorites(hwndDlg,gOutputPath);
					}
				}
				break;
			case IDCANCEL:
				// Cancel scanning if we are scanning.
				if( gScanning )
				{
					gCancelled = TRUE;
				}
				break;
			default:
				break;
            } 
		case WM_KEYDOWN:
			break;
    } 
    return FALSE; 
}

// update dialog, returns, TRUE if should continue, FALSE if user cancelled.
BOOL UpdateDialog( HWND hwndDlg, LPCTSTR nameStr, LPCTSTR urlStr)
{
	HWND dh;
	
	dh = GetDlgItem(hwndDlg,IDC_NAME);
	if( dh )
		SetWindowText(dh,nameStr);

	dh = GetDlgItem(hwndDlg,IDC_LINK);
	if( dh )
		SetWindowText(dh,urlStr);

	// Now do event handling so application doesn't freeze while processing.
	MSG msg;

	// process one message, if available, but don't wait on one.
	if( PeekMessage(&msg,NULL,0,0,PM_REMOVE) )
	{ 
		if (!IsWindow(hwndDlg) || !IsDialogMessage(hwndDlg, &msg)) 
		{ 
			TranslateMessage(&msg); 
			DispatchMessage(&msg); 
		}

		if( gCancelled )
			return FALSE;			
	}


	return TRUE;
}

// process favorites.
void ProcessFavorites(HWND hwndDlg, const char *outputPath )
{
	HRESULT hr;
	IMalloc *pMalloc = NULL;
	LPITEMIDLIST pidlFavorites = NULL;
	IShellFolder *psfDesktop = NULL;
	IShellFolder *psfFavorites = NULL;
	HWND dh;
	ofstream file;

	// flag that we are scanning so dialog proc won't be confused when we call it for progress.
	gScanning = TRUE;
	gCancelled = FALSE;
	
	// set string for output
	dh = GetDlgItem(hwndDlg,IDC_FILE);
	if( dh )
		SetWindowText(dh,outputPath);
	
	// enable cancel button
	dh = GetDlgItem(hwndDlg,IDCANCEL);
	if( dh )
		EnableWindow(dh,TRUE);

	// open output file.
	try
	{
		file.open(outputPath);
	}
	catch(...)
	{
		MessageBox(hwndDlg,"Error opening output file","Error",MB_OK | MB_ICONERROR );
	}

	if( file.is_open() )
	{
		// need access to the IMalloc interface to free other objects.
		hr = SHGetMalloc(&pMalloc);
		if( SUCCEEDED(hr) )
			// find favorites folder
			hr = SHGetSpecialFolderLocation(NULL, CSIDL_FAVORITES, &pidlFavorites);
		

		// get access to the desktop
		if( SUCCEEDED(hr) )
			hr = SHGetDesktopFolder(&psfDesktop);

		// bind to the favorites folder from the desktop object.
		if( SUCCEEDED(hr) )
			hr = psfDesktop->BindToObject(pidlFavorites, NULL, IID_IShellFolder,
				(LPVOID*)&psfFavorites);

		// output start of HTML
		file << "<html>\n";
		file << "<head><title>Favorites</title></head>\n";
		file << "<body>\n";
		file << "<h2>Favorites</h2>\n";
		file << "<ul>\n";

		// recursively process folders.
		if( SUCCEEDED(hr) )
			hr = ProcessFavFolder( hwndDlg, pMalloc,psfFavorites,file);	

		if( FAILED(hr) )
		{
			//  put up messagebox and tell user about error
			MessageBox(hwndDlg,"Error processing favorites","Error",MB_OK | MB_ICONERROR );
		}

		// output last of HTML
		file << "</ul>\n";
		file << "</body>\n";
		file << "</html>\n";


		// release the malloc object
		if( pMalloc )
			pMalloc->Release();


		try
		{
			file.close();

			/// if they cancelled, delete the output file.
			if( gCancelled )
			{
				unlink(outputPath);
			}

		}
		catch(...)
		{
			MessageBox(hwndDlg,"Error closing output file","Error",MB_OK | MB_ICONERROR );
		}



	}
	gScanning = FALSE;

	// disable cancel button
	dh = GetDlgItem(hwndDlg,IDCANCEL);
	if( dh )
		EnableWindow(dh,FALSE);

}


// free result with delete[]
LPTSTR MyStrRetToStr(LPITEMIDLIST item, STRRET *sret )
{
	USES_CONVERSION;

	// instead of StrRetToStr which requires IE 5
	int len;
	TCHAR *ts,*copy;
	
    switch (sret->uType)
    { 
	case STRRET_CSTR:
			len = strlen(sret->cStr);
			ts = sret->cStr;
			break;

		case STRRET_OFFSET:
			ts = ((char *)item) + sret->uOffset;
			break;

		case STRRET_WSTR:
			ts = OLE2T(sret->pOleStr);
			break;
    }
	if( ts )
	{
		len = _tcslen(ts);
		copy = new TCHAR[len+1];
		if( copy )
			memcpy(copy,ts,sizeof(TCHAR)*(len+1));
	}

	return copy;
}

HRESULT ProcessFavFolder(HWND hwndDlg, IMalloc *pMalloc, IShellFolder* psfFolder, ostream& file)
{
	HRESULT hr = S_OK;
	IEnumIDList *psfEnumList = NULL;
	LPITEMIDLIST item = NULL;
	LPCITEMIDLIST citem = NULL;
	STRRET nameRet, fileRet;
	LPTSTR nameStr = NULL, fileStr = NULL, urlStr = NULL;
	SHFILEINFO fileInfo;
	IShellFolder *psfSubFolder;

	hr = psfFolder->EnumObjects(NULL,
		SHCONTF_FOLDERS | SHCONTF_NONFOLDERS, 
		&psfEnumList );

	if( FAILED(hr) )
		goto cleanup;
	
	while( (hr=psfEnumList->Next(1,&item,NULL)) == NOERROR )
	{
		// Handle the item.
		hr = psfFolder->GetDisplayNameOf( item, SHGDN_NORMAL, &nameRet );
		if( FAILED(hr) ) 
			break;// break out and fall into cleanup

		hr = psfFolder->GetDisplayNameOf( item, SHGDN_FORPARSING, &fileRet );
		if( FAILED(hr) ) 
			break;// break out and fall into cleanup

		
		// convert them to normal strings			
		nameStr = MyStrRetToStr(item, &nameRet);
		fileStr = MyStrRetToStr(item, &fileRet);
		if( !nameStr || !fileStr )
		{
			// set hr to something meaningful and break
			hr = E_OUTOFMEMORY;
			break;
		}

		SHGetFileInfo(fileStr, NULL, &fileInfo, sizeof(fileInfo), SHGFI_TYPENAME | SHGFI_ATTRIBUTES );
		// if it is an actual folder, recurse
		if( (fileInfo.dwAttributes & SFGAO_FOLDER) && strcmp(fileInfo.szTypeName, _T("Channel Shortcut")) )
		{
			hr = psfFolder->BindToObject(item, NULL, IID_IShellFolder, (void **)&psfSubFolder);
			if( FAILED(hr) )
				break;

			// output a heading and begin a new list
			file << "<li><b>" << nameStr << "</b><ul>\n";

			hr = ProcessFavFolder(hwndDlg, pMalloc,psfSubFolder,file);
			if( FAILED(hr) )
				break;

			file << "</ul></li>\n";

			// release the subfolder
			psfSubFolder->Release();
		}
		else
		{
			// we are only interesting in actual URL's, not channels or links
			if  (!strcmp(fileInfo.szTypeName, _T("Internet Shortcut")))
			{
				// get the URL from the shortcut and process it.
				hr = GetURL(fileStr,&urlStr);
				if( FAILED(hr) )
					break;

				// Handle the URL
				if( !UpdateDialog(hwndDlg, nameStr,urlStr) )
					break; // user cancelled.

				// output a link to the file.
				file << "<li><a href=\"" << urlStr << "\">" << nameStr << "</a></li>\n";

			}
		}


		// free item and strings
		if( item )
			pMalloc->Free(item);

		if( nameStr )
			delete[] nameStr;
		if( fileStr )
			delete[] fileStr;

		// this one has to be freed with IMalloc because it was allocated by a Shell function
		if( urlStr )
			pMalloc->Free(urlStr);

		item = NULL;
		nameStr = NULL;
		fileStr = NULL;
		urlStr = NULL;

	}

cleanup:
	
	// free item and strings if needed
	if( item )
		pMalloc->Free(item);
	if( nameStr )
		pMalloc->Free(nameStr);
	if( fileStr )
		pMalloc->Free(fileStr);

	if( psfEnumList )
		psfEnumList->Release();

	return hr;
}


// get a URL from a URL shortcut.
HRESULT GetURL(const TCHAR *filePathStr, TCHAR **outURLStr)
{
	HRESULT hr;
	WCHAR wsz[MAX_PATH];  
	IUniformResourceLocator *urlLink = NULL;
	IPersistFile *persist = NULL;

	// set return to null
	*outURLStr = NULL;


	// create a IUniformResourceLocator object to get the URL from the path with
	hr = CoCreateInstance(CLSID_InternetShortcut, NULL, CLSCTX_INPROC_SERVER,
		IID_IUniformResourceLocator, (void **)&urlLink);

	if (SUCCEEDED(hr))
	{
		// need the IID_IPersistFile interface
		hr = urlLink->QueryInterface(IID_IPersistFile, (void **)&persist);
		if (SUCCEEDED(hr))
		{
			// path has to be in unicode.
			MultiByteToWideChar(CP_ACP, 0, filePathStr, -1, wsz, MAX_PATH);

			// load the path
			hr = persist->Load(wsz, STGM_READ);

			// get the URL 
			if (SUCCEEDED(hr))
				hr = urlLink->GetURL(outURLStr);

			// release
			persist->Release();
		}

		// release the object.
		urlLink->Release();
	}

	// return resutl
	return hr;
}  
